"""
Implementation of the base HFSModel and SumModel classes, based on the syntax used in the original satlas

NOTE: THIS IS NOT FULLY BENCHMARKED/DEVELOPED SO BUGS MIGHT BE PRESENT, AND NOT ALL FUNCTIONALITIES OF THE ORIGINAL SATLAS ARE IMPLEMENTED

.. moduleauthor:: Bram van den Borne <bram.vandenborne@kuleuven.be>
"""

from __future__ import annotations

from typing import Tuple, Union

import lmfit as lm
import pandas as pd
from numpy.typing import ArrayLike

from .core import Fitter, Source
from .models import HFS, PiecewiseConstant, Polynomial


class HFSModel:
    """Initializes a hyperfine spectrum Model with the given hyperfine parameters.

    Parameters
    ----------
    I : float
        Integer or half-integer value of the nuclear spin
    J : ArrayLike
        A sequence of 2 spins, respectively the J value of the lower state
        and the J value of the higher state
    ABC : ArrayLike
        A sequence of 2 A, 2 B and 2 C values, respectively for the lower and the higher state
    centroid : float, optional
        The centroid of the spectrum, by default 0
    fwhm : ArrayLike, length = 2, optional
        First element: The Gaussian FWHM of the Voigt profile, by default 50
        Second element: The Lorentzian FWHM of the Voigt profile, by default 50
    scale : float, optional
        The amplitude of the entire spectrum, by default 1.0
    background_params: ArrayLike, optional
        The coefficients of the polynomial background, by default [0.001]
    shape : str, optional
        Voigt only
    use_racah : bool, optional
        Use individual amplitudes are setting the Racah intensities, by default True
    use_saturation : bool, optional
        False only
    saturation: float, optional
        No saturation
    sidepeak_params : dict, optional
        keys:
    N : int, optional
        Number of sidepeaks to be generated, by default None
    Poisson : float, optional
        The poisson factor for the sidepeaks, by default 0
    Offset : float, optional
        Offset in units of x for the sidepeak, by default 0
    prefunc : callable, optional
        Transformation to be applied on the input before evaluation, by default None
    """

    def __init__(
        self,
        I: float,
        J: ArrayLike[float, float],
        ABC: ArrayLike[float, float, float, float, float, float],
        centroid: float = 0,
        fwhm: ArrayLike[float, float] = [50.0, 50.0],
        scale: float = 1.0,
        background_params: ArrayLike = [0.001],
        shape: str = "voigt",
        use_racah: bool = True,
        use_saturation: bool = False,
        saturation: float = 0.001,
        sidepeak_params: dict = {"N": None, "Poisson": 0, "Offset": 0},
        crystalballparams=None,
        pseudovoigtparams=None,
        asymmetryparams=None,
        name: str = "HFModel__",
    ):
        super(HFSModel, self).__init__()
        self.background_params = background_params
        if shape != "voigt":
            raise NotImplementedError("Only Voigt shape is supported.")
        if (
            crystalballparams != None
            or pseudovoigtparams != None
            or asymmetryparams != None
        ):
            raise NotImplementedError("Only Voigt shape is supported.")
        if name == "HFModel__":
            self.name = name + str(I).replace(".", "_")
        self.hfs = HFS(
            I,
            J,
            A=ABC[:2],
            B=ABC[2:4],
            C=ABC[4:6],
            scale=scale,
            df=centroid,
            fwhmg=fwhm[0],
            fwhml=fwhm[1],
            name=self.name,
            racah=use_racah,
            N=sidepeak_params["N"],
            offset=sidepeak_params["Offset"],
            poisson=sidepeak_params["Poisson"],
            prefunc=None,
        )
        self.params = self.hfs.params

    def set_expr(self, constraints: dict) -> None:
        """Set the expression to be used for the given parameters.
        The constraint should be a dict with following structure:

        key: string
            Parameter to constrain
        value: ArrayLike, length = 2
            First element: Factor to multiply
            Second element: Parameter that the key should be constrained to. {'Au':['0.5','Al']} results in Au = 0.5*Al
        """
        for cons in constraints.keys():
            self.hfs.params[
                cons
            ].expr = f"{constraints[cons][0]}*Fit___{self.name}___{constraints[cons][1]}"

    def fix_ratio(self, value, target="upper", parameter="A"):
        raise NotImplementedError("Use HFSModel.set_expr(...)")

    def set_variation(self, varyDict: dict) -> None:
        """Sets the variation of the fitparameters as supplied in the
        dictionary.

        Parameters
        ----------
        varyDict: dictionary
            A dictionary containing 'key: True/False' mappings with
            the parameter names as keys."""

        for p in varyDict.keys():
            self.hfs.params[p].vary = False

    def f(self, x: ArrayLike) -> ArrayLike:
        """Calculate the response for an unshifted spectrum with background

        Parameters
        ----------
        x : ArrayLike

        Returns
        -------
        ArrayLike
        """
        return self.hfs.fUnshifted(x) + Polynomial(
            self.background_params, name="bkg"
        ).f(x)

    def __call__(self, x: ArrayLike) -> ArrayLike:
        """Calculate the response for an unshifted spectrum with background

        Parameters
        ----------
        x : ArrayLike

        Returns
        -------
        ArrayLike
        """
        return self.hfs.fUnshifted(x) + Polynomial(
            self.background_params, name="bkg"
        ).f(x)

    def chisquare_fit(
        self,
        x: ArrayLike,
        y: ArrayLike,
        yerr: Union[ArrayLike, callable] = None,
        xerr: ArrayLike = None,
        func: callable = None,
        verbose: bool = False,
        hessian: bool = False,
        method: str = "leastsq",
        show_correl: bool = True,
    ) -> Tuple[bool, str]:
        """Perform a fit of this model to the data provided in this function.

        Parameters
        ----------
        x : ArrayLike
            x-values of the data points
        y : ArrayLike
            y-values of the data-points
        yerr : Union[ArrayLike, callable], optional
            1-sigma error on the y-values, values or function to be used on f(x),
            by default None=sqrt(f(x))
        xerr : ArrayLike, optional
            1-sigma error on the x-values, by default None
        func : callable, optional
            Not implemented
        verbose : bool, optional
            Not implemented, by default False
        hessian : bool, optional
            Not implemented, by default False
        method : str, optional
            Selects the method used by the :func:`lmfit.minimizer`, by default 'leastsq'

        Returns
        -------
        Tuple[bool, str]
            Returns a boolean indicating the success of the fit, and the message accompanying it.

        Raises
        ------
        NotImplementedError
            When the chosen options for func, verbose and Hessian result is not implemented.
        """
        if show_correl:
            print(
                "define whether you want to see the correlations in display_chisquare_fit(...)"
            )
        if func is not None:
            yerr = func
        datasource = Source(x, y, yerr=yerr, name="Fit")
        datasource.addModel(self.hfs)
        bkg = Polynomial(self.background_params, name="bkg")
        datasource.addModel(bkg)
        self.fitter = Fitter()
        self.fitter.addSource(datasource)
        self.fitter.fit(method=method)
        self.background_params = [
            list(bkg.params.values())[i].value
            for i in range(len(list(bkg.params.values())))
        ]
        return self.fitter.result.success, self.fitter.result.message

    def display_chisquare_fit(self, scaled: bool = True, **kwargs):
        """Generate a report of the fitting results.

        The report contains the best-fit values for the parameters and their uncertainties and correlations.

        Parameters
        ----------
        scaled: bool, optional
            Whether the errors are scaled with reduced chisquared, by default True, and only True
        show_correl : bool, optional
            Whether to show a list of sorted correlations, by default False
        min_correl : float, optional
            Smallest correlation in absolute value to show, by default 0.1

        Returns
        -------
        str
            Multi-line text of fit report.
        """
        if not scaled:
            raise NotImplementedError("Not implemented")
        return lm.fit_report(self.fitter.result, **kwargs)

    def get_result(
        self, selection: str = "chisquare"
    ) -> Tuple[list, list, list]:
        """Return the variable names, values and estimated error bars for the
        parameters as seperate lists.

        Parameters
        ----------
        selection: string, optional
            Selects if the chisquare ('chisquare' or 'any') or MLE values are
            used. Defaults to 'chisquare', and chisquare only

        Returns
        -------
        names, values, uncertainties: tuple of lists
            Returns a 3-tuple of lists containing the names of the parameters,
            the values and the estimated uncertainties, scaled with the reduced chisquared.
        """
        lmparamdict = self.fitter.pars["Fit"][self.name]
        return (
            list(lmparamdict.keys()),
            [
                lmparamdict[param_name].value
                for param_name in lmparamdict.keys()
            ],
            [lmparamdict[param_name].unc for param_name in lmparamdict.keys()],
        )

    def get_result_dict(
        self, method: str = "chisquare", scaled: bool = True
    ) -> dict:
        """Returns the fitted parameters in a dictionary of the form {name: [value, uncertainty]}.

        Parameters
        ----------
        method: {'chisquare', 'mle'}
            Selects which parameters have to be returned, by default 'chisquare', and only 'chisquare'
        scaled: boolean
            Selects if, in case of chisquare parameters, the uncertainty
            has to be scaled by sqrt(reduced_chisquare). Defaults to True, and only True

        Returns
        -------
        dict
            Dictionary of the form described above."""
        if (method.lower(), scaled) != ("chisquare", True):
            raise NotImplementedError("Not implemented")
        lmparamdict = self.fitter.pars["Fit"][self.name]
        return_dict = {
            param_name: [
                lmparamdict[param_name].value,
                lmparamdict[param_name].unc,
            ]
            for param_name in lmparamdict.keys()
        }
        return return_dict

    def get_result_frame(
        self,
        method: str = "chisquare",
        selected: bool = False,
        bounds: bool = False,
        vary: bool = False,
        scaled: bool = True,
    ) -> pd.DataFrame:
        """Returns the data from the fit in a pandas DataFrame.

        Parameters
        ----------
        method: str, optional
            Selects which fitresults have to be loaded. Can be 'chisquare' or
            'mle'. Defaults to 'chisquare', and only 'chisquare'.
        selected: list of strings, optional
            Selects the parameters that have any string in the list
            as a substring in their name. Set to *None* to select
            all parameters. Defaults to None, and only None.
        bounds: boolean, optional
            Selects if the boundary also has to be given. Defaults to
            False, and onlyb False.
        vary: boolean, optional
            Selects if only the parameters that have been varied have to
            be supplied. Defaults to False, and only False.
        scaled: boolean, optional
            Sets the uncertainty scaling with the reduced chisquare value. Default to True, and only True

        Returns
        -------
        resultframe: DataFrame
            Dateframe with MultiIndex, using the variable names as main column names
            and the two rows under for the value and the uncertainty"""
        result_dict = self.get_result_dict(method=method, scaled=scaled)
        return_frame = pd.DataFrame.from_dict(result_dict)
        return return_frame


class SumModel:
    """Initializes a hyperfine spectrum for the sum of multiple Models with the given models and a step background.

    Parameters
    ----------
    models : ArrayLike, with instances of HFSModel as elements
        The models that should be summed
    background_params: Dict with keys: 'values' and 'bounds' and values ArrayLike
        The bounds where the background changes stepwise in key 'bounds'
        The background values between the bounds
        i.e. {'values': [2,5], 'bounds':[-10]} means a background of 2 from -inf to -10, and a background of 5 from -10 to +inf
    name : string, optional
        Name of this summodel
    source_name : string, optional
        Name of the DataSource instance (from satlas2)
    """

    def __init__(
        self,
        models: list,
        background_params: list,
        name: str = "sum",
        source_name: str = "source",
    ):
        super(SumModel, self).__init__()
        self.name = name
        self.models = models
        self.background_params = background_params
        self._set_params()

    def _set_params(self):
        """Set the parameters of the underlying Models
        based on a large Parameters object
        """
        for model in self.models:
            try:
                p.add_many(*model.params.values())
            except:
                p = model.params.copy()
        self.params = p

    def set_variation(self, varyDict: dict):
        raise NotImplementedError("Do this at the HFSModel level")

    def f(self, x: ArrayLike) -> ArrayLike:
        """Calculate the response for a spectrum

        Parameters
        ----------
        x : ArrayLike

        Returns
        -------
        ArrayLike
        """
        for model in self.models:
            try:
                f += model.f(x)
            except UnboundLocalError:
                f = model.f(x)
        return f

    def __call__(self, x):
        return self.f(x)

    def chisquare_fit(
        self,
        x: ArrayLike,
        y: ArrayLike,
        yerr: Union[ArrayLike, callable] = None,
        xerr: ArrayLike = None,
        func: callable = None,
        verbose: bool = False,
        hessian: bool = False,
        method: str = "leastsq",
    ) -> Tuple[bool, str]:
        """Perform a fit of this model to the data provided in this function.

        Parameters
        ----------
        x : ArrayLike
            x-values of the data points
        y : ArrayLike
            y-values of the data-points
        yerr : Union[ArrayLike, callable], optional
            1-sigma error on the y-values, values or function to be used on f(x),
            by default None=sqrt(f(x))
        xerr : ArrayLike, optional
            1-sigma error on the x-values, by default None
        func : callable, optional
            Not implemented
        verbose : bool, optional
            Not implemented, by default False
        hessian : bool, optional
            Not implemented, by default False
        method : str, optional
            Selects the method used by the :func:`lmfit.minimizer`, by default 'leastsq'

        Returns
        -------
        Tuple[bool, str]
            Returns a boolean indicating the success of the fit, and the message accompanying it.

        Raises
        ------
        NotImplementedError
            When the chosen options for func, verbose and Hessian result is not implemented.
        """
        if (func, verbose, hessian) != (None, False, False):
            raise NotImplementedError("Not implemented")
        datasource = Source(x, y, yerr=yerr, name="Fit")
        for model in self.models:
            datasource.addModel(model)
        step_bkg = PiecewiseConstant(
            self.background_params["values"],
            self.background_params["bounds"],
            name="bkg",
        )
        self.models.append(step_bkg)
        datasource.addModel(step_bkg)
        self.fitter = Fitter()
        self.fitter.addSource(datasource)
        self.fitter.fit(method=method)
        return self.fitter.result.success, self.fitter.result.message

    def display_chisquare_fit(self, scaled: bool = True, **kwargs) -> str:
        """Generate a report of the fitting results.

        The report contains the best-fit values for the parameters and their uncertainties and correlations.

        Parameters
        ----------
        scaled: bool, optional
            Whether the errors are scaled with reduced chisquared, by default True, and only True
        show_correl : bool, optional
            Whether to show a list of sorted correlations, by default False
        min_correl : float, optional
            Smallest correlation in absolute value to show, by default 0.1

        Returns
        -------
        str
            Multi-line text of fit report.
        """
        if not scaled:
            raise NotImplementedError("Not implemented")
        return lm.fit_report(self.fitter.result, **kwargs)

    def get_result(
        self, selection: str = "chisquare"
    ) -> Tuple[list, list, list]:
        """Return the variable names, values and estimated error bars for the
        parameters as seperate lists.

        Parameters
        ----------
        selection: string, optional
            Selects if the chisquare ('chisquare' or 'any') or MLE values are
            used. Defaults to 'chisquare', and chisquare only

        Returns
        -------
        names, values, uncertainties: tuple of lists
            Returns a 3-tuple of lists containing the names of the parameters. The first list each tuple element contains the names/values/uncertainties of the first model added to the summodel, etc.
            The last list in each tuple element contains the names/values/uncertainties for the step background
            The values and the estimated uncertainties are always scaled with the reduced chisquared.
        """
        varnames = []
        varvalues = []
        varunc = []
        for model in self.models:
            lmparamdict = self.fitter.pars["Fit"][model.name]
            varnames.append(list(lmparamdict.keys()))
            varvalues.append(
                [
                    lmparamdict[param_name].value
                    for param_name in lmparamdict.keys()
                ]
            )
            varunc.append(
                [
                    lmparamdict[param_name].unc
                    for param_name in lmparamdict.keys()
                ]
            )
        return varnames, varvalues, varunc

    def get_result_dict(
        self, method: str = "chisquare", scaled: bool = True
    ) -> dict:
        """Returns the fitted parameters in a dictionary of the form {name of model in summodel : {name: [value, uncertainty]}}. Background values are under key 'bkg' in dictionary.

        Parameters
        ----------
        method: {'chisquare', 'mle'}
            Selects which parameters have to be returned, by default 'chisquare', and only 'chisquare'
        scaled: boolean
            Selects if, in case of chisquare parameters, the uncertainty
            has to be scaled by sqrt(reduced_chisquare). Defaults to True, and only True

        Returns
        -------
        dict
            Dictionary of the form described above."""
        if (method.lower(), scaled) != ("chisquare", True):
            raise NotImplementedError("Not implemented")
        return_dict = dict()
        for model in self.models:
            lmparamdict = self.fitter.pars["Fit"][model.name]
            return_dict[model.name] = {
                param_name: [
                    lmparamdict[param_name].value,
                    lmparamdict[param_name].unc,
                ]
                for param_name in lmparamdict.keys()
            }
        return return_dict

    def get_result_frame(
        self,
        method: str = "chisquare",
        selected: bool = False,
        bounds: bool = False,
        vary: bool = False,
        scaled: bool = True,
    ) -> pd.DataFrame:
        """Returns the data from the fit in a pandas DataFrame.

        Parameters
        ----------
        method: str, optional
            Selects which fitresults have to be loaded. Can be 'chisquare' or
            'mle'. Defaults to 'chisquare', and only 'chisquare'.
        selected: list of strings, optional
            Selects the parameters that have any string in the list
            as a substring in their name. Set to *None* to select
            all parameters. Defaults to None, and only None.
        bounds: boolean, optional
            Selects if the boundary also has to be given. Defaults to
            False, and onlyb False.
        vary: boolean, optional
            Selects if only the parameters that have been varied have to
            be supplied. Defaults to False, and only False.
        scaled: boolean, optional
            Sets the uncertainty scaling with the reduced chisquare value. Default to True, and only True

        Returns
        -------
        resultframe: DataFrame
            Dateframe with MultiIndex, using the model name + variable names as main column names
            and the two rows under for the value and the uncertainty"""
        result_dict = self.get_result_dict(method=method, scaled=scaled)
        return_frame = pd.DataFrame.from_dict(result_dict[self.models[0].name])
        return_frame = return_frame.add_prefix(f"{self.models[0].name}_")
        for model in self.models[1:]:
            df_to_add = pd.DataFrame.from_dict(result_dict[model.name])
            df_to_add = df_to_add.add_prefix(f"{model.name}_")
            return_frame = pd.concat([return_frame, df_to_add], axis=1)
        return return_frame


def chisquare_fit(
    model: Union["HFSModel", "SumModel"],
    x: ArrayLike,
    y: ArrayLike,
    yerr: Union[ArrayLike, callable],
    xerr: ArrayLike = None,
    method: str = "leastsq",
):
    """Perform a fit of the provided model to the data provided in this function.

    Parameters
    ----------
    model :
    x : ArrayLike
        x-values of the data points
    y : ArrayLike
        y-values of the data points
    yerr : ArrayLike
        1-sigma error on the y-values
    xerr : ArrayLike, optional
        1-sigma error on the x-values
    method : str, optional
        Selects the method used by the :func:`lmfit.minimizer`, by default 'leastsq'.
    show_correl : bool, optional
        Show correlations between fitted parameters in fit message, by default True

    Returns
    -------
    Instance of Fitter"""
    return model.chisquare_fit(x=x, y=y, yerr=yerr, xerr=xerr, method=method)



"""
Implementation of the HFSModel class, currently only supplied with a Voigt profile.

.. moduleauthor:: Wouter Gins <wouter.gins@kuleuven.be>
"""
from __future__ import annotations

from typing import Tuple

import numpy as np
import uncertainties as unc
from numpy.typing import ArrayLike
from scipy.special import voigt_profile, erf
from sympy.physics.wigner import wigner_3j, wigner_6j

from ..core import Model, Parameter

__all__ = ["HFS"]

sqrt2 = 2**0.5
sqrt2log2t2 = 2 * np.sqrt(2 * np.log(2))
log2 = np.log(2)


class HFS(Model):
    """Initializes a hyperfine spectrum Model with the given hyperfine parameters.

    Parameters
    ----------
    I : float
        Integer or half-integer value of the nuclear spin
    J : ArrayLike
        A sequence of 2 spins, respectively the J value of the lower state
        and the J value of the higher state
    A : ArrayLike, optional
        A sequence of 2 A values, respectively for the lower and the higher state, by default [0, 0]
    B : ArrayLike, optional
        A sequence of 2 B values, respectively for the lower and the higher state, by default [0, 0]
    C : ArrayLike, optional
        A sequence of 2 C values, respectively for the lower and the higher state, by default [0, 0]
    df : float, optional
        The centroid of the spectrum, by default 0
    fwhmg : float, optional
        The Gaussian FWHM of the Voigt profile, by default 50
    fwhml : float, optional
        The Lorentzian FWHM of the Voigt profile, by default 50
    name : str, optional
        Name of the model, by default 'HFS'
    peak: str, optional
        peak function to use, by default 'voigt'
    peak_kwargs: dict, optional
        additional fitting parameters for skew and custom peaks
    N : int, optional
        Number of sidepeaks to be generated, by default None
    offset : float, optional
        Offset in units of x for the sidepeak, by default 0
    poisson : float, optional
        The poisson factor for the sidepeaks, by default 0
    scale : float, optional
        The amplitude of the entire spectrum, by default 1.0
    racah : bool, optional
        Use individual amplitudes are setting the Racah intensities, by default True
    prefunc : callable, optional
        Transformation to be applied on the input before evaluation, by default None
        """
    def __init__(self,
                 I: float,
                 J: ArrayLike,
                 A: ArrayLike = [0, 0],
                 B: ArrayLike = [0, 0],
                 C: ArrayLike = [0, 0],
                 df: float = 0,
                 fwhmg: float = 50,
                 fwhml: float = 50,
                 name: str = 'HFS',
                 peak: str = 'voigt',
                 peak_kwargs: dict = None,
                 N: int = None,
                 offset: float = 0,
                 poisson: float = 0,
                 scale: float = 1.0,
                 racah: bool = True,
                 prefunc: callable = None):
        super().__init__(name, prefunc=prefunc)
        J1, J2 = J
        lower_F = np.arange(abs(I - J1), I + J1 + 1, 1)
        upper_F = np.arange(abs(I - J2), I + J2 + 1, 1)

        self.peakfunc = {
            'voigt': self.voigtPeak,
            'gaussian': self.gaussPeak,
            'lorentzian': self.lorentzPeak,
            'skewvoigt': self.skewPeak,
            'custom': self.customPeak
        }[peak.lower()]

        self.lines = []
        self.intensities = {}
        self.scaling_Al = {}
        self.scaling_Bl = {}
        self.scaling_Cl = {}
        self.scaling_Au = {}
        self.scaling_Bu = {}
        self.scaling_Cu = {}

        for i, F1 in enumerate(lower_F):
            for j, F2 in enumerate(upper_F):
                if abs(F2 - F1) <= 1 and not F2 == F1 == 0.0:
                    if F1 % 1 == 0:
                        F1_str = "{:.0f}".format(F1)
                    else:
                        F1_str = "{:.0f}_2".format(2 * F1)

                    if F2 % 1 == 0:
                        F2_str = "{:.0f}".format(F2)
                    else:
                        F2_str = "{:.0f}_2".format(2 * F2)

                    line = "{}to{}".format(F1_str, F2_str)
                    self.lines.append(line)

                    C1, D1, E1 = self.calcShift(I, J1, F1)
                    C2, D2, E2 = self.calcShift(I, J2, F2)

                    self.scaling_Al[line] = C1
                    self.scaling_Bl[line] = D1
                    self.scaling_Cl[line] = E1
                    self.scaling_Au[line] = C2
                    self.scaling_Bu[line] = D2
                    self.scaling_Cu[line] = E2

                    intens = float(
                        (2 * F1 + 1)
                        * (2 * F2 + 1)
                        * wigner_6j(J2, float(F2), I, float(F1), J1, 1.0) ** 2
                    )  # DO NOT REMOVE CAST TO FLOAT!!!
                    self.intensities["Amp" + line] = Parameter(
                        value=intens, min=0, vary=not racah
                    )

        norm = max([p.value for p in self.intensities.values()])
        for n, v in self.intensities.items():
            v.value /= norm

        pars = {
            "centroid": Parameter(value=df),
            "Al": Parameter(value=A[0]),
            "Au": Parameter(value=A[1]),
            "Bl": Parameter(value=B[0]),
            "Bu": Parameter(value=B[1]),
            "Cl": Parameter(value=C[0]),
            "Cu": Parameter(value=C[1]),
            "FWHMG": Parameter(value=fwhmg, min=0.01),
            "FWHML": Parameter(value=fwhml, min=0.01),
            "scale": Parameter(value=scale, min=0, vary=racah),
        }

        if peak.lower() == 'lorentzian':
            pars['FWHMG'].value,pars['FWHMG'].vary, pars['FWHMG'].min=0,False,0
        if peak.lower() == 'gaussian':
            pars['FWHML'].value,pars['FWHML'].vary,pars['FWHML'].min=0,False,0
        if peak_kwargs is not None:
            for peak_arg in peak_kwargs:
                pars[peak_arg] = Parameter(value=peak_kwargs[peak_arg]['value'], 
                                           min=peak_kwargs[peak_arg].get('min', -np.inf), 
                                           max=peak_kwargs[peak_arg].get('max', np.inf), 
                                           vary=peak_kwargs[peak_arg].get('vary', True),
                                           expr=peak_kwargs[peak_arg].get('expr', None))
        if N is not None:
            pars["N"] = Parameter(value=N, vary=False)
            pars["Offset"] = Parameter(value=offset)
            pars["Poisson"] = Parameter(value=poisson, min=0, max=1)
            self.f = self.fShifted
        else:
            self.f = self.fUnshifted
        pars = {**pars, **self.intensities}

        self.params = pars

        if I < 1.5 or J1 < 1.5:
            self.params["Cl"].vary = False
        if I < 1.5 or J2 < 1.5:
            self.params["Cu"].vary = False
        if I < 1 or J1 < 1:
            self.params["Bl"].vary = False
        if I < 1 or J2 < 1:
            self.params["Bu"].vary = False
        if I == 0 or J1 == 0:
            self.params["Al"].vary = False
        if I == 0 or J2 == 0:
            self.params["Au"].vary = False

    def fUnshifted(self, x: ArrayLike) -> ArrayLike:
        """:meta private:
        Calculate the response for an unshifted spectrum

        Parameters
        ----------
        x : ArrayLike

        Returns
        -------
        ArrayLike
        """
        centroid = self.params['centroid'].value
        Al = self.params['Al'].value
        Au = self.params['Au'].value
        Bl = self.params['Bl'].value
        Bu = self.params['Bu'].value
        Cl = self.params['Cl'].value
        Cu = self.params['Cu'].value
        scale = self.params['scale'].value

        try:
            result = np.zeros(len(x))
        except TypeError:
            x = np.array([x])
            result = np.zeros(len(x))
        x = self.transform(x)
        for line in self.lines:
            pos = centroid + Au * self.scaling_Au[line] + Bu * self.scaling_Bu[
                line] + Cu * self.scaling_Cu[line] - Al * self.scaling_Al[
                    line] - Bl * self.scaling_Bl[line] - Cl * self.scaling_Cl[
                        line]
            result += scale * self.params['Amp' + line].value * self.peak(
                x - pos)

        return result

    def fShifted(self, x: ArrayLike) -> ArrayLike:
        """:meta private:
        Calculate the response with :attr:`N` sidepeaks with an offset
        of :attr:`offset`

        Parameters
        ----------
        x : ArrayLike

        Returns
        -------
        ArrayLike
        """
        centroid = self.params['centroid'].value
        Al = self.params['Al'].value
        Au = self.params['Au'].value
        Bl = self.params['Bl'].value
        Bu = self.params['Bu'].value
        Cl = self.params['Cl'].value
        Cu = self.params['Cu'].value
        scale = self.params['scale'].value
        N = self.params['N'].value
        offset = self.params['Offset'].value
        poisson = self.params['Poisson'].value

        result = np.zeros(len(x))
        x = self.transform(x)
        for line in self.lines:
            pos = (
                centroid
                + Au * self.scaling_Au[line]
                + Bu * self.scaling_Bu[line]
                + Cu * self.scaling_Cu[line]
                - Al * self.scaling_Al[line]
                - Bl * self.scaling_Bl[line]
                - Cl * self.scaling_Cl[line]
            )
            for i in range(N + 1):
                result += self.params['Amp' + line].value * self.peak(
                    self.transform(x - i * offset) - pos) * (poisson**i) / np.math.factorial(i)
            result *= scale

        return result

    def peak(self, x: ArrayLike) -> ArrayLike:
        """:meta private:
        Calculates the profile given the peak_func method

        Parameters
        ----------
        x : ArrayLike
            Evaluation points

        Returns
        -------
        ArrayLike
        """
        returnvalue = self.peakfunc(x)
        return returnvalue

    def voigtPeak(self, x: ArrayLike) -> ArrayLike:
        """:meta private:
        Calculates the Voigt profile with the Gaussian
        and Lorentzian FWHM

        Parameters
        ----------
        x : ArrayLike
            Evaluation points

        Returns
        -------
        ArrayLike
        """
        sigma, gamma = self.params['FWHMG'].value / sqrt2log2t2, self.params['FWHML'].value / 2
        return voigt_profile(x, sigma, gamma) / voigt_profile(0, sigma, gamma)

    def lorentzPeak(self, x: ArrayLike) -> ArrayLike:
        """:meta private:
        Calculates the lorentzian profile 

        Parameters
        ----------
        x : ArrayLike
            Evaluation points

        Returns
        -------
        ArrayLike
        """
        gamma = self.params['FWHML'].value / 2
        return voigt_profile(x, 0, gamma) / voigt_profile(0, 0, gamma)

    def gaussPeak(self, x: ArrayLike) -> ArrayLike:
        """:meta private:
        Calculates the Gaussian profile

        Parameters
        ----------
        x : ArrayLike
            Evaluation points

        Returns
        -------
        ArrayLike
        """
        sigma = self.params['FWHMG'].value / sqrt2log2t2
        return voigt_profile(x, sigma, 0) / voigt_profile(0, sigma, 0)

    def skewPeak(self, x: ArrayLike) -> ArrayLike:
        """:meta private:
        Calculates a skewed voigt profile with gaussian and lorentzian FWHM and a skewness parameter

        Parameters
        ----------
        x : ArrayLike
            Evaluation points

        Returns
        -------
        ArrayLike
        """
        sigma, gamma = self.params['FWHMG'].value / 2 * np.sqrt(2 * np.log(2)), self.params['FWHML'].value/2
        erf_x = self.params['skew'].value*x/self.params['FWHMG'].value
        return (voigt_profile(x, sigma, gamma) / voigt_profile(0, sigma, gamma))*(1+erf(erf_x/np.sqrt(2)))

    def customPeak(self, x: ArrayLike) -> ArrayLike:
        """:meta private:
        Calculate a custom peak

        Parameters
        ----------
        x : ArrayLike
            Evaluation points

        Returns
        -------
        ArrayLike
        """
        raise NotImplementedError

    def calcShift(self, I: float, J: float, F: int) -> ArrayLike:
        """:meta private:
        Calculate the coefficients for the energy shift due to the hyperfine
        interaction up to the octupole moment. A general equation is used
        so extending to higher orders is possible.

        Parameters
        ----------
        I : float
            Nuclear spin
        J : float
            Electronic spin
        F : int
            Hyperfine level spin

        Returns
        -------
        ArrayLike
            Individual coefficients, in ascending order
        """
        phase = (-1) ** (I + J + F)
        contrib = []
        for k in range(1, 4):
            n = float(wigner_6j(I, J, float(F), J, I, k))
            d = float(
                wigner_3j(I, k, I, -I, 0, I) * wigner_3j(J, k, J, -J, 0, J)
            )
            shift = phase * n / d
            if not np.isfinite(shift):
                contrib.append(0)
            else:
                if k == 1:
                    shift = shift * (I * J)
                elif k == 2:
                    shift = shift / 4
                contrib.append(shift)
        return contrib

    def pos(self) -> ArrayLike:
        """Returns the positions of the peaks in MHz in the hyperfine spectrum

        Returns
        -------
        ArrayLike
        """
        centroid = self.params["centroid"].value
        Al = self.params["Al"].value
        Au = self.params["Au"].value
        Bl = self.params["Bl"].value
        Bu = self.params["Bu"].value
        Cl = self.params["Cl"].value
        Cu = self.params["Cu"].value
        pos = []
        for line in self.lines:
            p = (
                centroid
                + Au * self.scaling_Au[line]
                + Bu * self.scaling_Bu[line]
                + Cu * self.scaling_Cu[line]
                - Al * self.scaling_Al[line]
                - Bl * self.scaling_Bl[line]
                - Cl * self.scaling_Cl[line]
            )
            pos.append(p)
        return pos

    def calculateFWHM(self) -> Tuple[float, float]:
        """Calculate the total FWHM of the profiles, with uncertainty,
        taking the correlations into account.

        Returns
        -------
        Tuple[float, float]
            Tuple of the form (value, uncertainty)
        """
        G, Gu = self.params["FWHMG"].value, self.params["FWHMG"].unc
        L, Lu = self.params["FWHML"].value, self.params["FWHML"].unc
        try:
            correl = self.params["FWHMG"].correl["FWHML"]
        except KeyError:
            correl = 0
        G, L = unc.correlated_values_norm(
            [(G, Gu), (L, Lu)], np.array([[1, correl], [correl, 1]])
        )
        fwhm = 0.5346 * L + (0.2166 * L * L + G * G) ** 0.5
        return fwhm.nominal_value, fwhm.std_dev



"""
Implementation of various functions that ease the work, but do not belong in one of the other modules.

.. moduleauthor:: Wouter Gins <wouter.gins@kuleuven.be>
"""
from __future__ import annotations

from typing import Optional, Tuple, Union

import numpy as np
from numpy.typing import ArrayLike
from scipy.stats import chi2, norm, poisson

from .core import Model

__all__ = ["weightedAverage", "poissonInterval", "generateSpectrum"]


def weightedAverage(
    x: ArrayLike, sigma: ArrayLike, axis: Optional[int] = None
) -> Tuple[float, float]:
    r"""Takes the weighted average of an array of values and the associated
    errors. Calculates the scatter and statistical error, and returns
    the greater of these two values.

    Parameters
    ----------
    x: ArrayLike
        Array-like assortment of measured values, is transformed into a
        1D-array.
    sigma: ArrayLike
        Array-like assortment of errors on the measured values, is transformed
        into a 1D-array.
    axis: Optional[int]
        Axis over which the weighted average should be calculated

    Returns
    -------
    Tuple[float, float]
        Returns a tuple (weighted average, uncertainty), with the uncertainty
        being the greater of the uncertainty calculated from the statistical
        uncertainty and the scattering uncertainty.

    Note
    ----
    The formulas used are

    .. math::

        \left\langle x\right\rangle_{weighted} &= \frac{\sum_{i=1}^N \frac{x_i}
                                                                 {\sigma_i^2}}
                                                      {\sum_{i=1}^N \frac{1}
                                                                {\sigma_i^2}}

        \sigma_{stat}^2 &= \frac{1}{\sum_{i=1}^N \frac{1}{\sigma_i^2}}

        \sigma_{scatter}^2 &= \frac{\sum_{i=1}^N \left(\frac{x_i-\left\langle
                                                    x\right\rangle_{weighted}}
                                                      {\sigma_i}\right)^2}
               {\left(N-1\right)\sum_{i=1}^N \frac{1}{\sigma_i^2}}"""
    x = np.array(x)
    sigma = np.array(sigma)
    Xstat = (1 / sigma**2).sum(axis=axis)
    Xm = (x / sigma**2).sum(axis=axis) / Xstat
    Xscatt = (((x - Xm) / sigma) ** 2).sum(axis=axis) / ((len(x) - 1) * Xstat)
    Xstat = 1 / Xstat
    return Xm, np.maximum.reduce([Xstat, Xscatt], axis=axis) ** 0.5


def poissonInterval(
    data: ArrayLike,
    sigma: float = 1,
    alpha: Optional[float] = None,
    mean: bool = False,
) -> Tuple[float, float]:
    """Calculates the confidence interval
    for the mean of a Poisson distribution.

    Parameters
    ----------
    data: ArrayLike
        Samples of separate Poisson distributions.
    sigma: float
        The significance level given in equivalent sigma.
        Defaults to 1-sigma.
    alpha: Optional[float]
        Significance level of interval. If given, *sigma* is ignored.
    mean: bool
        Set to True if the exact mean is given, by default False

    Returns
    -------
    low, high: Tuple[float, float]
        Lower and higher limits for the interval."""
    if alpha is None:
        a = (1 - norm.cdf(np.abs(sigma))) * 2
    else:
        a = alpha
    if mean:
        a = 1 - a
        low, high = poisson.interval(a, data)
    else:
        low, high = (
            chi2.ppf(a / 2, 2 * data) / 2,
            chi2.ppf(1 - a / 2, 2 * data + 2) / 2,
        )
    low = np.nan_to_num(low)
    return low, high


def generateSpectrum(
    models: Union[Model, list],
    x: ArrayLike,
    generator: Optional[callable] = np.random.default_rng().poisson,
) -> ArrayLike:
    """Generates a dataset based on the models and x-values provided.

    Parameters
    ----------
    models : Union[Model, list]
        A single Model or list of models. In case of a list, all models are summed together.
    x : ArrayLike
        The x values for which a y value has to be generated.
    generator : callable, optional
        A callable with one parameter that returns a random value based on this.
        The default is a Poisson generator.

    Returns
    -------
    ArrayLike
        A same-sized array as x with values given by feeding the Model.f(x) value
        to the generator.
    """

    def evaluate(x):
        try:
            for model in models:
                try:
                    f += model.f(x)
                except UnboundLocalError:
                    f = model.f(x)
        except TypeError:
            f = models.f(x)
        return f

    y = evaluate(x)
    y = generator(y)
    return y



"""
Reimplementation of several features in the emcee and lmfit packages, in order to make it work correctly.

.. moduleauthor:: Wouter Gins <wouter.gins@kuleuven.be>
"""
import multiprocessing

import emcee
import lmfit.minimizer
import numpy as np
from emcee.autocorr import AutocorrError
from lmfit import Minimizer

AbortFitException = lmfit.minimizer.AbortFitException
from typing import Dict, List, Union

try:
    import pandas as pd
    from pandas import isnull

    HAS_PANDAS = True
except ImportError:
    HAS_PANDAS = False
    isnull = np.isnan
try:
    import dill  # noqa: F401

    HAS_DILL = True
except ImportError:
    HAS_DILL = False
_make_random_gen = lmfit.minimizer._make_random_gen
isnull = lmfit.minimizer.isnull
_nan_policy = lmfit.minimizer._nan_policy

__all__ = ["SATLASSampler", "SATLASHDFBackend", "SATLASMinimizer", "minimize"]


def ndarray_to_list_of_dicts(
    x: np.ndarray, key_map: Dict[str, Union[int, List[int]]]
) -> List[Dict[str, Union[np.number, np.ndarray]]]:
    """
    A helper function to convert a ``np.ndarray`` into a list
    of dictionaries of parameters. Used when parameters are named.
    Args:
      x (np.ndarray): parameter array of shape ``(N, n_dim)``, where
        ``N`` is an integer
      key_map (Dict[str, Union[int, List[int]]):
    Returns:
      list of dictionaries of parameters
    """
    return [{key: xi[val] for key, val in key_map.items()} for xi in x]


class SATLASSampler(emcee.EnsembleSampler):
    def compute_log_prob(self, coords):
        """Calculate the vector of log-probability for the walkers

        Args:
            coords: (ndarray[..., ndim]) The position vector in parameter
                space where the probability should be calculated.

        This method returns:

        * log_prob: A vector of log-probabilities with one entry for each
          walker in this sub-ensemble.
        * blob: The list of meta data returned by the ``log_post_fn`` at
          this position or ``None`` if nothing was returned.

        """
        p = coords

        # Check that the parameters are in physical ranges.
        if np.any(np.isinf(p)):
            raise ValueError("At least one parameter value was infinite")
        if np.any(np.isnan(p)):
            raise ValueError("At least one parameter value was NaN")

        # If the parmaeters are named, then switch to dictionaries
        if self.params_are_named:
            p = ndarray_to_list_of_dicts(p, self.parameter_names)

        # Run the log-probability calculations (optionally in parallel).
        if self.vectorize:
            results = self.log_prob_fn(p)
        else:
            # If the `pool` property of the sampler has been set (i.e. we want
            # to use `multiprocessing`), use the `pool`'s map method.
            # Otherwise, just use the built-in `map` function.
            if self.pool is not None:
                map_func = self.pool.map
            else:
                map_func = map
            results = list(map_func(self.log_prob_fn, p))

        log_prob = np.array([float(l) for l in results])
        blob = None

        # Check for log_prob returning NaN.
        if np.any(np.isnan(log_prob)):
            raise ValueError("Probability function returned NaN")

        return log_prob, blob


class SATLASHDFBackend(emcee.backends.HDFBackend):
    @property
    def labels(self):
        with self.open() as f:
            g = f[self.name]
            return g.attrs["labels"]

    @labels.setter
    def labels(self, labels):
        with self.open("a") as f:
            g = f[self.name]
            g.attrs["labels"] = labels


class SATLASMinimizer(Minimizer):
    def process_walk(self, params, chain):
        result = self.prepare_fit(params)
        params = result.params
        nvarys = result.nvarys
        result.method = "emcee"

        flatchain = chain.reshape((-1, nvarys))
        steps = chain.shape[0]
        nwalkers = chain.shape[1]
        quantiles = np.percentile(flatchain, [15.87, 50, 84.13], axis=0)

        for i, var_name in enumerate(result.var_names):
            std_l, median, std_u = quantiles[:, i]
            params[var_name].value = median
            params[var_name].stderr = 0.5 * (std_u - std_l)
            params[var_name].correl = {}

        params.update_constraints()

        # work out correlation coefficients
        corrcoefs = np.corrcoef(flatchain.T)

        for i, var_name in enumerate(result.var_names):
            for j, var_name2 in enumerate(result.var_names):
                if i != j:
                    result.params[var_name].correl[var_name2] = corrcoefs[i, j]

        result.errorbars = True
        result.nvarys = len(result.var_names)
        result.nfev = nwalkers * steps

        try:
            result.acor = emcee.autocorr.integrated_time(chain)
        except AutocorrError as e:
            print(str(e))
            result.acor = emcee.autocorr.integrated_time(chain, tol=0)
        return result

        # Calculate the residual with the "best fit" parameters

    def emcee(
        self,
        params=None,
        steps=1000,
        nwalkers=100,
        burn=0,
        thin=1,
        ntemps=1,
        load=False,
        convergence=False,
        convergence_iter=50,
        convergence_tau=0.01,
        pos=None,
        reuse_sampler=False,
        workers=1,
        float_behavior="posterior",
        is_weighted=True,
        seed=None,
        progress=True,
        mcmc_kwargs={},
        sampler_kwargs={},
        sampler=emcee.EnsembleSampler,
    ):
        if ntemps > 1:
            msg = (
                "'ntemps' has no effect anymore, since the PTSampler was "
                "removed from emcee version 3."
            )
            raise DeprecationWarning(msg)

        tparams = params
        # if you're reusing the sampler then nwalkers have to be
        # determined from the previous sampling
        if reuse_sampler:
            if not hasattr(self, "sampler") or not hasattr(self, "_lastpos"):
                raise ValueError(
                    "You wanted to use an existing sampler, but "
                    "it hasn't been created yet"
                )
            if len(self._lastpos.shape) == 2:
                nwalkers = self._lastpos.shape[0]
            elif len(self._lastpos.shape) == 3:
                nwalkers = self._lastpos.shape[1]
            tparams = None

        result = self.prepare_fit(params=tparams)
        params = result.params

        # check if the userfcn returns a vector of residuals
        out = self.userfcn(params, *self.userargs, **self.userkws)
        out = np.asarray(out).ravel()
        if out.size > 1 and is_weighted is False:
            # we need to marginalise over a constant data uncertainty
            if "__lnsigma" not in params:
                # __lnsigma should already be in params if is_weighted was
                # previously set to True.
                params.add(
                    "__lnsigma", value=0.01, min=-np.inf, max=np.inf, vary=True
                )
                # have to re-prepare the fit
                result = self.prepare_fit(params)
                params = result.params

        result.method = "emcee"

        # Removing internal parameter scaling. We could possibly keep it,
        # but I don't know how this affects the emcee sampling.
        bounds = []
        var_arr = np.zeros(len(result.var_names))
        i = 0
        for par in params:
            param = params[par]
            if param.expr is not None:
                param.vary = False
            if param.vary:
                var_arr[i] = param.value
                i += 1
            else:
                # don't want to append bounds if they're not being varied.
                continue

            param.from_internal = lambda val: val
            lb, ub = param.min, param.max
            if lb is None or lb is np.nan:
                lb = -np.inf
            if ub is None or ub is np.nan:
                ub = np.inf
            bounds.append((lb, ub))
        bounds = np.array(bounds)

        self.nvarys = len(result.var_names)

        # set up multiprocessing options for the samplers
        auto_pool = None
        # sampler_kwargs = {}
        if isinstance(workers, int) and workers > 1 and HAS_DILL:
            auto_pool = multiprocessing.Pool(workers)
            sampler_kwargs["pool"] = auto_pool
        elif hasattr(workers, "map"):
            sampler_kwargs["pool"] = workers

        # function arguments for the log-probability functions
        # these values are sent to the log-probability functions by the sampler.
        lnprob_args = (self.userfcn, params, result.var_names, bounds)
        lnprob_kwargs = {
            "is_weighted": is_weighted,
            "float_behavior": float_behavior,
            "userargs": self.userargs,
            "userkws": self.userkws,
            "nan_policy": self.nan_policy,
        }

        sampler_kwargs["args"] = lnprob_args
        sampler_kwargs["kwargs"] = lnprob_kwargs

        # set up the random number generator
        rng = _make_random_gen(seed)

        backend = sampler_kwargs.pop("backend")
        if backend is not None:
            if not load:
                backend.reset(nwalkers, self.nvarys)
            else:
                nwalkers = backend.shape[0]
            backend.labels = result.var_names
        sampler_kwargs["backend"] = backend
        # now initialise the samplers

        if load:
            p0 = None
        else:
            p0 = 1 + rng.randn(nwalkers, self.nvarys) * 1.0e-4
            p0 *= var_arr
        sampler_kwargs["pool"] = auto_pool
        self.sampler = sampler(
            nwalkers, self.nvarys, self._lnprob, **sampler_kwargs
        )

        # user supplies an initialisation position for the chain
        # If you try to run the sampler with p0 of a wrong size then you'll get
        # a ValueError. Note, you can't initialise with a position if you are
        # reusing the sampler.
        if pos is not None and not reuse_sampler:
            tpos = np.asfarray(pos)
            if p0.shape == tpos.shape:
                pass
            # trying to initialise with a previous chain
            elif tpos.shape[-1] == self.nvarys:
                tpos = tpos[-1]
            else:
                raise ValueError("pos should have shape (nwalkers, nvarys)")
            p0 = tpos

        # if you specified a seed then you also need to seed the sampler
        if seed is not None:
            self.sampler.random_state = rng.get_state()

        # now do a production run, sampling all the time
        try:
            output = None
            old_tau = np.inf
            check = int(np.ceil(1000 / nwalkers))
            converged = False
            if p0 is None:
                p0 = self.sampler._previous_state
            for output in self.sampler.sample(
                p0, iterations=steps, progress=progress, **mcmc_kwargs
            ):
                if convergence:
                    if self.sampler.iteration % check:
                        continue
                    tau = self.sampler.get_autocorr_time(tol=0)
                    converged = np.all(
                        tau * convergence_iter < self.sampler.iteration
                    )
                    converged &= np.all(
                        np.abs(old_tau - tau) / tau < convergence_tau
                    )
                    if converged:
                        break
                    old_tau = tau
            if converged:
                print("emcee stopped due to convergence")
            self._lastpos = output.coords
        except AbortFitException:
            result.aborted = True
            result.message = (
                "Fit aborted by user callback. Could not estimate error-bars."
            )
            result.success = False
            result.nfev = self.result.nfev
            output = None

        # discard the burn samples and thin
        chain = self.sampler.get_chain(thin=thin, discard=burn)[..., :, :]
        lnprobability = self.sampler.get_log_prob(thin=thin, discard=burn)[
            ..., :
        ]
        flatchain = chain.reshape((-1, self.nvarys))
        if not result.aborted:
            quantiles = np.percentile(flatchain, [15.87, 50, 84.13], axis=0)

            for i, var_name in enumerate(result.var_names):
                std_l, median, std_u = quantiles[:, i]
                params[var_name].value = median
                params[var_name].stderr = 0.5 * (std_u - std_l)
                params[var_name].correl = {}

            params.update_constraints()

            # work out correlation coefficients
            corrcoefs = np.corrcoef(flatchain.T)

            for i, var_name in enumerate(result.var_names):
                for j, var_name2 in enumerate(result.var_names):
                    if i != j:
                        result.params[var_name].correl[var_name2] = corrcoefs[
                            i, j
                        ]

        result.chain = np.copy(chain)
        result.lnprob = np.copy(lnprobability)
        result.errorbars = True
        result.nvarys = len(result.var_names)
        result.nfev = nwalkers * steps

        try:
            result.acor = self.sampler.get_autocorr_time()
        except AutocorrError as e:
            print(str(e))
            result.acor = self.sampler.get_autocorr_time(tol=0, quiet=True)
        result.acceptance_fraction = self.sampler.acceptance_fraction

        # Calculate the residual with the "best fit" parameters
        out = self.userfcn(params, *self.userargs, **self.userkws)
        result.residual = _nan_policy(
            out, nan_policy=self.nan_policy, handle_inf=False
        )

        # If uncertainty was automatically estimated, weight the residual properly
        if (not is_weighted) and (result.residual.size > 1):
            if "__lnsigma" in params:
                result.residual = result.residual / np.exp(
                    params["__lnsigma"].value
                )

        # Calculate statistics for the two standard cases:
        if isinstance(result.residual, np.ndarray) or (
            float_behavior == "chi2"
        ):
            result._calculate_statistics()

        # Handle special case unique to emcee:
        # This should eventually be moved into result._calculate_statistics.
        elif float_behavior == "posterior":
            result.ndata = 1
            result.nfree = 1

            # assuming prior prob = 1, this is true
            _neg2_log_likel = -2 * result.residual

            # assumes that residual is properly weighted, avoid overflowing np.exp()
            result.chisqr = np.exp(min(650, _neg2_log_likel))

            result.redchi = result.chisqr / result.nfree
            result.aic = _neg2_log_likel + 2 * result.nvarys
            result.bic = _neg2_log_likel + np.log(result.ndata) * result.nvarys

        if auto_pool is not None:
            auto_pool.terminate()

        return result


def minimize(
    fcn,
    params,
    method="leastsq",
    args=None,
    kws=None,
    iter_cb=None,
    scale_covar=True,
    nan_policy="raise",
    reduce_fcn=None,
    calc_covar=True,
    max_nfev=None,
    **fit_kws,
):
    minimizer_kws = fit_kws.pop("minimizer_kws", {})
    fitter = SATLASMinimizer(
        fcn,
        params,
        fcn_args=args,
        fcn_kws=kws,
        iter_cb=iter_cb,
        scale_covar=scale_covar,
        nan_policy=nan_policy,
        reduce_fcn=reduce_fcn,
        calc_covar=calc_covar,
        max_nfev=max_nfev,
        **fit_kws,
    )
    return fitter.minimize(method=method, **minimizer_kws)



"""
Functions for the generation of plots related to the fitting results.

.. moduleauthor:: Wouter Gins <wouter.gins@kuleuven.be>
"""
import copy
from typing import List, Optional, Tuple

import matplotlib as mpl
import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np
import tqdm
import uncertainties as u
from scipy import optimize
from scipy.stats import chi2

from .core import Fitter
from .overwrite import SATLASHDFBackend

inv_color_list = [
    "#7acfff",
    "#fff466",
    "#00c48f",
    "#ff8626",
    "#ff9cd3",
    "#0093e6",
]
color_list = [c for c in reversed(inv_color_list)]
cmap = mpl.colors.ListedColormap(color_list)
cmap.set_over(color_list[-1])
cmap.set_under(color_list[0])
invcmap = mpl.colors.ListedColormap(inv_color_list)
invcmap.set_over(inv_color_list[-1])
invcmap.set_under(inv_color_list[0])

__all__ = [
    "generateChisquareMap",
    "generateCorrelationPlot",
    "generateWalkPlot",
]


def _make_axes_grid(
    no_variables,
    width=6,
    height=6,
    cbar=True,
    left=0.1,
    right=0.9,
    top=0.9,
    bottom=0.1,
):
    """Makes a triangular grid of axes, with a colorbar axis next to it.

    Parameters
    ----------
    no_variables: int
        Number of variables for which to generate a figure.
    padding: float
        Padding around the figure (in cm).
    cbar_size: float
        Width of the colorbar (in cm).
    axis_padding: float
        Padding between axes (in cm).

    Returns
    -------
    fig, axes, cbar: tuple
        Tuple containing the figure, a 2D-array of axes and the colorbar axis.
    """

    fig = plt.figure(constrained_layout=True, figsize=(width, height))
    width_ratios = [1] * no_variables
    if cbar:
        width_ratios.extend([0.1] * 2)
    gs = gridspec.GridSpec(
        nrows=no_variables,
        ncols=no_variables + 2 * cbar,
        left=left,
        right=right,
        top=top,
        bottom=bottom,
        wspace=0,
        hspace=0,
        figure=fig,
        width_ratios=width_ratios,
    )

    # Pre-allocate a 2D-array to hold the axes.
    axes = np.array(
        [[None for _ in range(no_variables)] for _ in range(no_variables)],
        dtype="object",
    )

    for i, I in zip(range(no_variables), reversed(range(no_variables))):
        for j in reversed(range(no_variables)):
            # Only create axes on the lower triangle.
            if I + j < no_variables:
                # Share the x-axis with the plot on the diagonal,
                # directly above the plot.
                sharex = axes[j, j] if i != j else None
                # Share the y-axis among the 2D maps along one row,
                # but not the plot on the diagonal!
                sharey = axes[i, i - 1] if (i != j and i - 1 != j) else None
                a = fig.add_subplot(gs[i, j], sharex=sharex, sharey=sharey)
                a.label_outer()
                plt.setp(a.xaxis.get_majorticklabels(), rotation=45)
                plt.setp(a.yaxis.get_majorticklabels(), rotation=45)
            else:
                a = None
            if i == j:
                a.set_yticks([])
                a.set_yticklabels([])
            axes[i, j] = a

    axes = np.array(axes)
    if cbar:
        cbar = fig.add_subplot(gs[:, -1])
    else:
        cbar = None
    return fig, axes, cbar


def generateChisquareMap(
    fitter: Fitter,
    filter: Optional[List[str]] = None,
    method: str = "chisquare",
    resolution_diag: int = 15,
    resolution_map: int = 15,
    fit_kws: dict = {},
    source: bool = False,
    model: bool = True,
):
    """:meta private:
    Generates a correlation map for either the chisquare or the MLE method.
    On the diagonal, the chisquare or loglikelihood is drawn as a function of one fixed parameter.
    Refitting to the data each time gives the points on the line. A dashed line is drawn on these
    plots, with the intersection with the plots giving the correct confidence interval for the
    parameter. In solid lines, the interval estimated by the fitting routine is drawn.
    On the offdiagonal, two parameters are fixed and the model is again fitted to the data.
    The change in chisquare/loglikelihood is mapped to 1, 2 and 3 sigma contourmaps.

    Parameters
    ----------
    fitter: :class:`.Fitter`
        Fitter instance for which the chisquare map must be created.

    Other parameters
    ----------------
    filter: list of strings
        Only the parameters matching the names given in this list will be used
        to generate the maps.
    resolution_diag: int
        Number of points for the line plot on each diagonal.
    resolution_map: int
        Number of points along each dimension for the meshgrids.
    fit_kws: dictionary
        Dictionary of keywords to pass on to the fitting routine.
    npar: int
        Number of parameters for which simultaneous predictions need to be made.
        Influences the uncertainty estimates from the parabola."""

    title = "{}\n${}_{{-{}}}^{{+{}}}$"
    title_e = "{}\n$({}_{{-{}}}^{{+{}}})e{}$"

    try:
        orig_value = fitter.chisqr
    except AttributeError:
        fitter.fit(**fit_kws)
        orig_value = fitter.chisqr
    if method.lower().startswith("llh"):
        orig_value = fitter.llh_result
    result = copy.deepcopy(fitter.result)
    orig_params = copy.deepcopy(fitter.lmpars)

    ranges = {}

    param_names = []
    no_params = 0
    for p in orig_params:
        if orig_params[p].vary and (
            filter is None or any([f in p for f in filter])
        ):
            no_params += 1
            param_names.append(p)
    fig, axes, cbar = _make_axes_grid(
        no_params, axis_padding=0, cbar=no_params > 1
    )

    split_names = [name.split("___") for name in param_names]
    sources = [name[0] for name in split_names]
    models = [name[1] for name in split_names]
    var_names = [name[2] for name in split_names]
    to_be_combined = [var_names]
    if model:
        to_be_combined.insert(0, models)
    if source:
        to_be_combined.insert(0, sources)

    var_names = [" ".join(tbc) for tbc in zip(*to_be_combined)]

    # Make the plots on the diagonal: plot the chisquare/likelihood
    # for the best fitting values while setting one parameter to
    # a fixed value.
    saved_params = copy.deepcopy(fitter.lmpars)
    for i in range(no_params):
        params = copy.deepcopy(saved_params)
        ranges[param_names[i]] = {}

        # Set the y-ticklabels.
        ax = axes[i, i]
        ax.set_title(param_names[i])
        if i == no_params - 1:
            if method.lower().startswith("chisquare"):
                ax.set_ylabel(r"$\Delta\chi^2$")
            else:
                ax.set_ylabel(r"$\Delta\mathcal{L}$")
                fit_kws["llh_selected"] = True

        # Select starting point to determine error widths.
        value = orig_params[param_names[i]].value
        stderr = orig_params[param_names[i]].stderr
        stderr = stderr if stderr is not None else 0.01 * np.abs(value)
        stderr = stderr if stderr != 0 else 0.01 * np.abs(value)

        right = value + stderr
        left = value - stderr
        params[param_names[i]].vary = False

        ranges[param_names[i]]["left_val"] = 3 * left - 2 * value
        ranges[param_names[i]]["right_val"] = 3 * right - 2 * value
        value_range = np.linspace(
            3 * left - 2 * value, right * 3 - 2 * value, resolution_diag
        )
        chisquare = np.zeros(len(value_range))
        # Calculate the new value, and store it in the array. Update the progressbar.
        # with tqdm.tqdm(value_range, desc=param_names[i], leave=True) as pbar:
        for j, v in enumerate(value_range):
            params[param_names[i]].value = v
            fitter.lmpars = params
            fitter.fit(prepFit=False, **fit_kws)
            if fitter.llh_result is not None:
                chisquare[j] = fitter.llh_result - orig_value
            else:
                chisquare[j] = fitter.chisqr - orig_value
                # pbar.update(1)
        # Plot the result
        ax.plot(value_range, chisquare, color="k")

        c = "#0093e6"
        ax.axvline(right, ls="dashed", color=c)
        ax.axvline(left, ls="dashed", color=c)
        ax.axvline(value, ls="dashed", color=c)
        up = "{:.2ug}".format(u.ufloat(value, stderr))
        down = "{:.2ug}".format(u.ufloat(value, stderr))
        val = up.split("+/-")[0].split("(")[-1]
        r = up.split("+/-")[1].split(")")[0]
        l = down.split("+/-")[1].split(")")[0]
        if "e" in up or "e" in down:
            ex = up.split("e")[-1]
            ax.set_title(title_e.format(var_names[i], val, l, r, ex))
        else:
            ax.set_title(title.format(var_names[i], val, l, r))
        # Restore the parameters.
        fitter.lmpars = orig_params

    for i, j in zip(*np.tril_indices_from(axes, -1)):
        params = copy.deepcopy(orig_params)
        ax = axes[i, j]
        x_name = param_names[j]
        y_name = param_names[i]
        if j == 0:
            ax.set_ylabel(var_names[i])
        if i == no_params - 1:
            ax.set_xlabel(var_names[j])
        right = ranges[x_name]["right_val"]
        left = ranges[x_name]["left_val"]
        x_range = np.linspace(left, right, resolution_map)

        right = ranges[y_name]["right_val"]
        left = ranges[y_name]["left_val"]
        y_range = np.linspace(left, right, resolution_map)

        X, Y = np.meshgrid(x_range, y_range)
        Z = np.zeros(X.shape)
        i_indices, j_indices = np.indices(Z.shape)
        params[param_names[i]].vary = False
        params[param_names[j]].vary = False

        for k, l in zip(i_indices.flatten(), j_indices.flatten()):
            x = X[k, l]
            y = Y[k, l]
            params[param_names[j]].value = x
            params[param_names[i]].value = y
            fitter.lmpars = params
            fitter.fit(prepFit=False, **fit_kws)
            if fitter.llh_result is not None:
                Z[k, l] = (fitter.llh_result - orig_value) * 2
            else:
                Z[k, l] = fitter.chisqr - orig_value

        Z = -Z
        bounds = []
        for bound in [0.997300204, 0.954499736, 0.682689492]:
            chifunc = (
                lambda x: chi2.cdf(x, 1) - bound
            )  # Calculate 1 sigma boundary
            bounds.append(-optimize.root(chifunc, 1).x[0])
        bounds.append(0)
        bounds = np.array(bounds)
        norm = mpl.colors.BoundaryNorm(bounds, invcmap.N)
        contourset = ax.contourf(X, Y, Z, bounds, cmap=invcmap, norm=norm)
        fitter.lmpars = copy.deepcopy(orig_params)
    try:
        cbar = plt.colorbar(contourset, cax=cbar, orientation="vertical")
        cbar.ax.yaxis.set_ticks([-7.5, -4.5, -1.5])
        cbar.ax.set_yticklabels([r"3$\sigma$", r"2$\sigma$", r"1$\sigma$"])
    except:
        pass
    for a in axes.flatten():
        if a is not None:
            for label in a.get_xticklabels()[::2]:
                label.set_visible(False)
            for label in a.get_yticklabels()[::2]:
                label.set_visible(False)
    fitter.result = result
    fitter.updateInfo()
    return fig, axes, cbar


def generateCorrelationPlot(
    filename: str,
    filter: Optional[List[str]] = None,
    bins: Optional[int] = None,
    burnin: int = 0,
    thin: int = 1,
    autoprocess: bool = False,
    source: bool = True,
    model: bool = True,
    binreduction: int = 1,
    bin2dreduction: int = 1,
    progress: bool = False,
    width: float = 6,
    height: float = 6,
    left: float = 0.15,
    right: float = 0.95,
    top: float = 0.85,
    bottom: float = 0.15,
) -> Tuple[plt.Figure, Tuple[plt.Axes], plt.Axes]:
    """Given the random walk data, creates a triangle plot: distribution of
    a single parameter on the diagonal axes, 2D contour plots with 1, 2 and
    3 sigma contours on the off-diagonal. The 1-sigma limits based on the
    percentile method are also indicated, as well as added to the title.

    Parameters
    ----------
    filename : str
        Filename for the h5 file containing the data from the walk.
    filter : List[str], optional
        Only this list of columns is used for the plot, by default None.
    bins : int, optional
        Use this number of bins for the plotting.
        Applies the same number of bins for each parameter.
        If supplied as a list, length must match the number of
        parameters. By default None.
    burnin : int, optional
        Number of initial steps from the random walk to be discarded,
        by default 0.
    thin : int, optional
        Take only every ``thin`` steps from the chain. (default: ``1``)
    autoprocess : bool, optional
        Based on the autocorrelation time of the random walk, perform an
        automatic burn-in and thinning estimate, by default False.
    source : bool, optional
        Add the source name to the plot titles, by default True.
    model : bool, optional
        Add the model name to the plot titles, by default True.
    binreduction : int, optional
        Reduces the amount of bins in the 1D case by this factor,
        by default 1.
    bin2dreduction : int, optional
        Further reduces the amount of bins in the 2D case by this factor,
        by default 1.
    progress : bool, optional
        Show a progress bar of processing the parameters, by default False.
    width : float, optional
        Width in inches of the figure, by default 6
    height : float, optional
        Height in inches of the figure, by default 6
    Left : float, optional
        Extent of the left of the figure, in fraction, by default 0.15
    right : float, optional
        Extent of the right of the figure, in fraction, by default 0.95
    top : float, optional
        Extent of the top of the figure, in fraction, by default 0.85
    bottom : float, optional
        Extent of the bottom of the figure, in fraction, by default 0.15

    Returns
    -------
    Tuple[plt.Figure, Tuple[plt.Axes], plt.Axes]
        Tuple containing the figure, the individual axes, and the colorbar axis.

    Note
    ----
    When estimated automatically, the ``burnin`` and ``thin`` are set to
    respectively

    .. math::
        2\cdot\\textrm{max}\\left(\\tau\\right)

    and

    .. math::
        \\textrm{min}\\left(\\tau\\right)/2"""

    reader = SATLASHDFBackend(filename)
    var_names = list(reader.labels)
    split_names = [name.split("___") for name in var_names]
    sources = [name[0] for name in split_names]
    models = [name[1] for name in split_names]
    var_names = [name[2] for name in split_names]
    to_be_combined = [var_names]
    if model:
        to_be_combined.insert(0, models)
    if source:
        to_be_combined.insert(0, sources)

    var_names = ["\n".join(tbc) for tbc in zip(*to_be_combined)]
    full_names = list(reader.labels)

    if autoprocess:
        tau = reader.get_autocorr_time(tol=0)
        burnin = int(2 * np.max(tau))
        thin = int(0.5 * np.min(tau))
    data = reader.get_chain(flat=True, discard=burnin, thin=thin)

    if filter is not None:
        filter = [
            (c, n)
            for f in filter
            for (c, n) in zip(var_names, full_names)
            if f in c
        ]
    else:
        filter = list(zip(var_names, full_names))
    with tqdm.tqdm(
        total=len(filter) + (len(filter) ** 2 - len(filter)) / 2,
        leave=True,
        disable=not progress,
    ) as pbar:
        fig, axes, cbar = _make_axes_grid(
            len(filter),
            width=width,
            height=height,
            left=left,
            right=right,
            top=top,
            bottom=bottom,
        )
        fig.set_layout_engine(None)

        metadata = {}
        if not isinstance(bins, list):
            bins = [bins for _ in filter]
        for i, val in enumerate(filter):
            name, full_name = val
            pbar.set_description(name)
            ax = axes[i, i]
            bin_index = i
            i = full_names.index(full_name)
            x = data[:, i]

            if bins[bin_index] is None:
                width = (
                    3.5 * np.std(x) / x.size ** (1 / 3)
                )  # Scott's rule for binwidth
                bins[bin_index] = int(
                    min(int(x.ptp() / width), 1000) / binreduction
                )
            try:
                (
                    n,
                    b,
                    p,
                ) = ax.hist(
                    x, int(bins[bin_index]), histtype="step", color="k"
                )
            except TypeError:
                bins[bin_index] = 50
                (
                    n,
                    b,
                    p,
                ) = ax.hist(
                    x, int(bins[bin_index]), histtype="step", color="k"
                )
            q = [15.87, 50, 84.13]
            q16, q50, q84 = np.percentile(x, q)
            metadata[full_name] = {
                "bins": bins[bin_index],
                "min": x.min(),
                "max": x.max(),
            }

            title = "{}\n${}_{{-{}}}^{{+{}}}$"
            title_e = "{}\n$({}_{{-{}}}^{{+{}}})e{}$"
            up = "{:.2ug}".format(u.ufloat(q50, np.abs(q84 - q50)))
            down = "{:.2ug}".format(u.ufloat(q50, np.abs(q50 - q16)))
            param_val = up.split("+/-")[0].split("(")[-1]
            r = up.split("+/-")[1].split(")")[0]
            l = down.split("+/-")[1].split(")")[0]
            if "e" in up or "e" in down:
                ex = up.split("e")[-1]
                t = ax.set_title(title_e.format(name, param_val, l, r, ex))
            else:
                t = ax.set_title(title.format(name, param_val, l, r))

            qvalues = [q16, q50, q84]
            c = "#0093e6"
            for q in qvalues:
                ax.axvline(q, ls="dashed", color=c)
            pbar.update(1)

        for i, j in zip(*np.tril_indices_from(axes, -1)):
            x_name, x_fullname = filter[j]
            y_name, y_fullname = filter[i]
            pbar.set_description(", ".join([x_name, y_name]))
            ax = axes[i, j]
            if j == 0:
                ax.set_ylabel(y_name)
            if i == len(filter) - 1:
                ax.set_xlabel(x_name)
            j = full_names.index(x_fullname)
            i = full_names.index(y_fullname)
            x = data[:, j]
            y = data[:, i]
            x_min, x_max, x_bins = (
                metadata[x_fullname]["min"],
                metadata[x_fullname]["max"],
                metadata[x_fullname]["bins"],
            )
            y_min, y_max, y_bins = (
                metadata[y_fullname]["min"],
                metadata[y_fullname]["max"],
                metadata[y_fullname]["bins"],
            )
            X = np.linspace(x_min, x_max, int(x_bins / bin2dreduction) + 1)
            Y = np.linspace(y_min, y_max, int(y_bins / bin2dreduction) + 1)
            H, X, Y = np.histogram2d(
                x.flatten(), y.flatten(), bins=(X, Y), weights=None
            )
            X1, Y1 = 0.5 * (X[1:] + X[:-1]), 0.5 * (Y[1:] + Y[:-1])
            X, Y = X[:-1], Y[:-1]
            H = (H - H.min()) / (H.max() - H.min())

            Hflat = H.flatten()
            inds = np.argsort(Hflat)[::-1]
            Hflat = Hflat[inds]
            sm = np.cumsum(Hflat)
            sm /= sm[-1]
            levels = 1.0 - np.exp(-0.5 * np.arange(1, 3.1, 1) ** 2)
            V = np.empty(len(levels))
            for i, v0 in enumerate(levels):
                try:
                    V[i] = Hflat[sm <= v0][-1]
                except:
                    V[i] = Hflat[0]

            bounds = np.unique(np.concatenate([[H.max()], V])[::-1])
            norm = mpl.colors.BoundaryNorm(bounds, invcmap.N)

            contourset = ax.contourf(
                X1, Y1, H.T, bounds, cmap=invcmap, norm=norm
            )
            pbar.update(1)
        try:
            cbar = plt.colorbar(contourset, cax=cbar, orientation="vertical")
            ticks = cbar.ax.get_yticks()
            dfticks = (ticks[1:] - ticks[:-1]) / 2
            ticks = ticks[:-1] + dfticks
            cbar.ax.yaxis.set_ticks(ticks)
            cbar.ax.set_yticklabels([r"3$\sigma$", r"2$\sigma$", r"1$\sigma$"])
        except:
            cbar = None
    return fig, axes, cbar


def generateWalkPlot(
    filename: str,
    filter: Optional[List[str]] = None,
    burnin: int = 0,
    thin: int = 1,
    autoprocess: bool = False,
    source: bool = False,
    model: bool = True,
    progress: bool = False,
) -> Tuple[plt.Figure, Tuple[plt.Axes]]:
    """Given the random walk data, the random walk for the selected parameters
    is plotted.

    Parameters
    ----------
    filename : str
        Filename for the h5 file containing the data from the walk.
    filter : List[str], optional
        Only this list of columns is used for the plot, by default None.
    burnin : int, optional
        Number of initial steps from the random walk to be discarded,
        by default 0.
    thin : int, optional
        Take only every ``thin`` steps from the chain. (default: ``1``)
    autoprocess : bool, optional
        Based on the autocorrelation time of the random walk, perform an
        automatic burn-in and thinning estimate, by default False.
    source : bool, optional
        Add the source name to the plot titles, by default False.
    model : bool, optional
        Add the model name to the plot titles, by default True.
    progress : bool, optional
        Show a progress bar of processing the parameters, by default False.

    Returns
    -------
    Tuple[plt.Figure, Tuple[plt.Axes]]
        Tuple containing the figure and the individual axes.

    Note
    ----
    When estimated automatically, the ``burnin`` and ``thin`` are set to
    respectively

    .. math::
        2\cdot\\textrm{max}\\left(\\tau\\right)

    and

    .. math::
        \\textrm{min}\\left(\\tau\\right)/2"""
    reader = SATLASHDFBackend(filename)
    var_names = reader.labels
    split_names = [name.split("___") for name in var_names]
    sources = [name[0] for name in split_names]
    models = [name[1] for name in split_names]
    var_names = [name[2] for name in split_names]
    to_be_combined = [var_names]
    if model:
        to_be_combined.insert(0, models)
    if source:
        to_be_combined.insert(0, sources)

    var_names = ["\n".join(tbc) for tbc in zip(*to_be_combined)]
    full_names = list(reader.labels)

    if autoprocess:
        tau = reader.get_autocorr_time(tol=0)
        burnin = int(2 * np.max(tau))
        thin = int(0.5 * np.min(tau))
    data = reader.get_chain(flat=False, discard=burnin, thin=thin)
    plot_x = np.arange(data.shape[0]) * thin + burnin

    if filter is not None:
        filter = [
            (c, n)
            for f in filter
            for (c, n) in zip(var_names, full_names)
            if f in c
        ]
    else:
        filter = list(zip(var_names, full_names))
    with tqdm.tqdm(
        total=len(filter), leave=True, disable=not progress
    ) as pbar:
        fig = plt.figure(constrained_layout=True)
        gs = gridspec.GridSpec(nrows=len(filter), ncols=1, figure=fig)

        old_ax = None
        axes = []
        for i, val in enumerate(filter):
            ax = fig.add_subplot(gs[i, 0], sharex=old_ax)
            old_ax = ax
            ax.label_outer()
            name, full_name = val
            pbar.set_description(name)
            i = full_names.index(full_name)
            x = data[:, :, i]
            q50 = np.percentile(x, [50.0])
            ax.plot(plot_x, x, alpha=0.3, color="gray")
            ax.set_ylabel(name)
            ax.axhline(q50, color="k")
            axes.append(ax)
            pbar.update(1)
        ax.set_xlabel("Step")
    pbar.close()
    return fig, axes



from . import core, interface, models, overwrite, plotting, utilities
from .core import *
from .interface import *
from .models import *
from .overwrite import *
from .plotting import *
from .utilities import *

__version__ = "0.2.7"
